from functools import lru_cache
from typing import Tuple
from decimal import Decimal

import tiktoken
import replicate
from replicate.exceptions import ReplicateError
from tenacity import retry, stop_after_attempt, wait_random_exponential, retry_if_exception, after_log

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.settings import chat_settings
import logging

logger = logging.getLogger(__name__)


class ReplicateModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer

    @retry(reraise=True,
           stop=stop_after_attempt(chat_settings.retry_attempts),
           retry=retry_if_exception(lambda x: isinstance(x, ReplicateError) and len(x.args) > 0 and
                                              "Request was throttled." in x.args[0]),
           wait=wait_random_exponential(multiplier=chat_settings.retry_wait_multiplier,
                                 min=chat_settings.retry_wait_min,
                                 max=chat_settings.retry_wait_max),
           after=after_log(logger, logging.DEBUG) )
    def _call_api(self, prompt: str) -> str:

        response = replicate.run(self.spec.internal_name, input={
            "prompt": prompt,
            "system_prompt": self.spec.system_message if self.spec.system_message is not None else "",
            "top_k": 50,
            "top_p": 0.9,
            "temperature": 0,
            "max_new_tokens": 1024,
            "presence_penalty": 0,
            "frequency_penalty": 0
        })
        answer = "".join(response)
        return answer

    @lru_cache(maxsize=chat_settings.max_cache_size)
    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, num_tokens_input = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        response = self._call_api(final_prompt)
        tokens_response = len(get_tokens_encoded(response, self.encoding))
        cost_input = (int(num_tokens_input) * Decimal(self.spec.price_per_1k_tokens / 1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (int(tokens_response) * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        final_cost = cost_input + cost_output
        return response, final_cost



from decimal import Decimal
from typing import *

from tiktoken.core import Encoding

from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification


def get_tokens_encoded(prompt: str, encoding: Encoding) -> List[int]:
    return encoding.encode(prompt)


def truncate_prompt(prompt: Prompt, encoding: Encoding, max_tokens: int) -> Tuple[str, int]:
    tokens_history = get_tokens_encoded(prompt.history, encoding)
    tokens_instructions = get_tokens_encoded(prompt.instructions(), encoding)
    tokens_data = get_tokens_encoded(prompt.available_data_string(), encoding)
    if len(tokens_instructions) > max_tokens:
        raise Exception("instructions exceed token limit")
    num_tokens = len(tokens_history) + len(tokens_instructions) + len(tokens_data)
    if num_tokens > max_tokens:
        # check if instructions + data are fitting
        if len(tokens_instructions) + len(tokens_data) <= max_tokens:
            # cut the history only
            history_truncated = encoding.decode(tokens_history[0:(max_tokens-len(tokens_instructions)-len(tokens_data))])
            return f"{history_truncated} {prompt.instructions()} {prompt.available_data_string()}", max_tokens
        else:
            # cut the data and don't return history
            available_data_truncated = encoding.decode(tokens_data[0:(max_tokens-len(tokens_instructions))])
            return f"{prompt.instructions()} \n {available_data_truncated}", max_tokens
    else:
        return str(prompt), num_tokens


class LLMBaseModel:

    spec: MlModelSpecification

    def __init__(self, spec: MlModelSpecification):
        self.spec = spec

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.spec)

    def __eq__(self, other):
        if isinstance(other, LLMBaseModel):
            return self.spec == other.spec
        return NotImplemented


from typing import *
from functools import reduce
import json

from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.models.llm_models.structuring_schema import get_prompt_schema_item
from parsee.templates.general_structuring_schema import GeneralQueryItemSchema, StructuringItemSchema
from parsee.extraction.extractor_dataclasses import ParseeMeta, ExtractedSource, ParseeAnswer
from parsee.utils.enums import DocumentType, SearchStrategy
from parsee.storage.interfaces import StorageManager
from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl, StructuredTable


MAIN_QUESTION_STR = "main_question"


class GeneralQueriesPromptBuilder:

    storage: StorageManager

    def __init__(self, storage: StorageManager):
        self.storage = storage

    def format_single_item(self, answer: ParseeAnswer, item: GeneralQueryItemSchema, meta_items: List[StructuringItemSchema]) -> Dict:
        schema_item = get_prompt_schema_item(item)
        output = {MAIN_QUESTION_STR: schema_item.parsed_to_raw(answer.class_value)}
        for meta_item in meta_items:
            meta_schema_item = get_prompt_schema_item(meta_item)
            meta_answers = [x for x in answer.meta if x.class_id == meta_item.id]
            answer_chosen = meta_answers[0].class_value if len(meta_answers) > 0 else "n/a"
            output[meta_item.id] = meta_schema_item.parsed_to_raw(answer_chosen)
        return output

    def build_raw_value(self, answers: List[ParseeAnswer], item: GeneralQueryItemSchema, meta_items: List[StructuringItemSchema], add_sources: bool = True) -> str:
        if len(meta_items) > 0:
            output = {"answers": [self.format_single_item(x, item, meta_items) for x in answers]}
        else:
            # if no meta items, there can be only one answer
            output = self.format_single_item(answers[0], item, meta_items)
        if add_sources:
            output["sources"] = [x.element_index for x in answers[0].sources]
        return json.dumps(output)

    def get_relevant_elements(self, schema_item: GeneralQueryItemSchema, document: StandardDocumentFormat) -> List[ExtractedEl]:

        if schema_item.searchStrategy == SearchStrategy.VECTOR:
            return self.storage.vector_store.find_closest_elements(document, schema_item.title, schema_item.keywords, False)
        elif schema_item.searchStrategy == SearchStrategy.START:
            return document.elements
        else:
            raise NotImplementedError

    def get_elements_text(self, elements: List[ExtractedEl], document: StandardDocumentFormat):
        llm_text = "This is the available data to answer the question (in the following, if tables have empty cells, they are omitted, if a cell spans several columns, values might be repeated for each cell):\n"
        for el in elements:
            llm_text += f"[{el.source.element_index}]: {el.get_text_and_surrounding_elements_text(document.elements) if isinstance(el, StructuredTable) else el.get_text_llm(True)} [/{el.source.element_index}]\n"
        return llm_text

    def build_prompt(self, structuring_item: GeneralQueryItemSchema, meta_items: List[StructuringItemSchema], document: StandardDocumentFormat, relevant_elements: List[ExtractedEl], multimodal: bool = False, max_images: Optional[int] = None, max_image_size: Optional[int] = None) -> Prompt:

        if not multimodal:
            general_info = "You are supposed to answer a question based on text fragments that are provided. " \
                           "The fragments start with a number in square brackets and then the actual text. The end of the fragment is shown by the same number in square brackets, only that the number is preceded by a slash. E.g. [22] Some Text [/22]. The lower the number of the fragment, " \
                           "the earlier the fragment appeared in the document (reading from left to right, top to bottom). "
        else:
            general_info = "You are supposed to answer a question based on one or several images that are provided below."

        prompt_schema_item = get_prompt_schema_item(structuring_item)

        additional_info_str = f" Additional info: {structuring_item.additionalInfo}" if structuring_item.additionalInfo.strip() != "" else ""
        main_question = f'The question is: {structuring_item.title} {additional_info_str} {prompt_schema_item.get_possible_values_str()}'

        relevant_meta_items = [x for x in meta_items if structuring_item.metaInfoIds is not None and x.id in structuring_item.metaInfoIds]

        # build full example
        source_examples = [ExtractedSource(DocumentType.PDF, None, None, 241, None), ExtractedSource(DocumentType.PDF, None, None, 423, None)]
        meta_examples = [ParseeMeta("test", 0, source_examples, x.id, get_prompt_schema_item(x).get_example(True), 0.8) for x in relevant_meta_items]
        sample_answer = ParseeAnswer("sample", source_examples, structuring_item.id, prompt_schema_item.get_example(), "", True, meta_examples)
        example_output = self.build_raw_value([sample_answer], structuring_item, relevant_meta_items, not multimodal)
        sources_format = "Under the key 'sources', you should also provide the numbers of the text fragments you used to answer the question." if not multimodal else ""

        full_example = f"Your answer should be formatted as a JSON object, under the key '{MAIN_QUESTION_STR}' you can put as value the answer to the question in the designated format. {sources_format} For example your answer could look like this: {example_output}" \
            if len(relevant_meta_items) == 0 else f"Your answer should be formatted as a JSON object. Under the key 'answers' can be one or more JSON object(s), with key value pairs for the main question and the meta items. For the main question use the key '{MAIN_QUESTION_STR}', for the meta items use their ID as key. {sources_format} For example your answer could look like this: {example_output}"

        # add prompting for meta
        if len(relevant_meta_items) > 0:
            main_question += "\n We also want to retrieve some meta information. In the following we will present the meta item ID and then the additional question to be answered, format: (META_ID): QUESTION."
            for meta_item in relevant_meta_items:
                meta_prompt_item = get_prompt_schema_item(meta_item)
                main_question += f"\n({meta_item.id}): {meta_item.title} {meta_item.additionalInfo} {meta_prompt_item.get_possible_values_str()}"

        prompt = Prompt(general_info, main_question, None,
                        full_example,
                        self.get_elements_text(relevant_elements, document) if not multimodal else self.storage.image_creator.get_images(document, relevant_elements, max_images, max_image_size))
        return prompt



from typing import *
from dataclasses import dataclass
from decimal import Decimal
import os

from parsee.utils.enums import ModelType


@dataclass
class MlModelSpecification:
    name: str
    model_id: str
    internal_name: str
    model_type: ModelType
    file_path: Optional[str]
    price_per_1k_tokens: Optional[Decimal]
    price_per_1k_output_tokens: Optional[Decimal]
    price_per_image: Optional[Decimal]
    max_tokens: Optional[int]
    api_key: Optional[str]
    only_questions: Optional[List[str]]
    only_elements: Optional[List[str]]
    only_meta: Optional[List[str]]
    only_mappings: Optional[List[str]]
    stats: Optional[Dict]
    multimodal: bool
    max_images: Optional[int]
    max_image_pixels: Optional[int]
    max_output_tokens: Optional[int]
    system_message: Optional[str]

    def model_path(self) -> Union[None, str]:
        if self.file_path is None:
            return None
        return self.file_path

    def settings_path(self) -> Union[None, str]:
        if self.file_path is None:
            return None
        model_dir = os.path.dirname(self.file_path)
        return os.path.join(model_dir, "settings.pkl")

    def __key(self):
        return (self.name, self.model_id, self.internal_name, self.model_type, self.file_path, self.price_per_1k_tokens,
                self.price_per_1k_output_tokens, self.price_per_image, self.max_tokens, self.api_key,
                self.only_questions, self.only_elements, self.only_meta, self.only_mappings, self.stats,
                self.multimodal, self.max_images, self.max_image_pixels, self.max_output_tokens, self.system_message)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, MlModelSpecification):
            return self.__key() == other.__key()
        return NotImplemented


from typing import List, Dict

from parsee.extraction.extractor_elements import StandardDocumentFormat, ElementGroup, get_text_distance, FinalOutputTableColumn, FinalOutputTable
from parsee.extraction.extractor_dataclasses import ParseeMeta, ParseeLocation
from parsee.templates.job_template import JobTemplate
from parsee.utils.constants import *


def text_distance_table_groups(el1: ElementGroup, el2: ElementGroup, elements):
    el_idx1 = el1.base_el().source.element_index
    el_idx2 = el2.base_el().source.element_index
    indices_to_exclude = [x.source.element_index for x in el1.components] + [x.source.element_index for x in el2.components]

    return get_text_distance(el_idx1, el_idx2, elements, indices_to_exclude, True)


def assemble(job_template: JobTemplate, document: StandardDocumentFormat, locations: List[ParseeLocation]) -> List[ElementGroup]:

    detection_schema_by_class = {}
    for item in job_template.detection.items:
        detection_schema_by_class[item.id] = item
    
    # transform predictions slightly
    candidates_by_class: Dict[str, List[ElementGroup]] = {}
    for k, prediction in enumerate(locations):
        class_value = prediction.detected_class
        if class_value is not None:
            if class_value not in candidates_by_class:
                candidates_by_class[class_value] = []
            # find detection item
            candidates_by_class[class_value].append(ElementGroup(class_value, prediction, detection_schema_by_class[prediction.detected_class].collapseColumns))

    predictions_by_element_index = {}
    for pred in locations:
        predictions_by_element_index[pred.source.element_index] = pred

    # decide which partial matches to merge
    for class_value, structured_elements in candidates_by_class.items():

        if len(structured_elements) > 1:
            # check if el can be merged with previous
            for k in range(len(structured_elements) - 1, 0, -1):

                loc = structured_elements[k].base_el()
                merge_candidate = structured_elements[k - 1].closest_el(structured_elements[k])

                # both have to have quite high probability
                if loc.prob < MERGE_MIN_CONFIDENCE or loc.prob < MERGE_MIN_CONFIDENCE:
                    continue

                # both partial prob have to be above threshold
                if loc.partial_prob < PARTIAL_MIN_CONFIDENCE or merge_candidate.partial_prob < PARTIAL_MIN_CONFIDENCE:
                    continue

                # tables need to be uninterrupted by other candidates
                min_index = min(loc.source.element_index, merge_candidate.source.element_index)
                max_index = max(loc.source.element_index, merge_candidate.source.element_index)
                found_inbetween = False
                for el_index in range(min_index + 1, max_index):
                    if el_index in predictions_by_element_index and predictions_by_element_index[el_index].prob > THRESHOLD_INBETWEEN_MERGE:
                        found_inbetween = True
                        break
                if found_inbetween:
                    continue

                # there can't be too much text between elements
                td = get_text_distance(loc.source.element_index, merge_candidate.source.element_index, document.elements, include_tables=True)

                if td > TEXT_DISTANCE_MERGE_THRESHOLD and loc.prob < 2:
                    continue

                # merge elements
                structured_elements[k - 1].merge_with(structured_elements[k])
                # delete from list
                structured_elements.pop(k)

    # decide which statements to keep if more than 1 close together
    for class_value, structured_elements in candidates_by_class.items():
        if class_value in detection_schema_by_class and detection_schema_by_class[class_value].takeBestInProximity:
            if len(structured_elements) > 1:
                # if some elements are close together, take one with highest score
                distances = []
                to_del = []
                for k in range(0, len(structured_elements) - 1):
                    for kk in range(k + 1, len(structured_elements)):
                        distance = text_distance_table_groups(structured_elements[k], structured_elements[kk], document.elements)
                        distances.append((distance, k, kk))
                        if distance <= TEXT_DISTANCE_CLOSE_STATEMENT_DETECTION:
                            # delete statement with lower probability
                            idx_to_delete = k if structured_elements[k].prob_combined() < structured_elements[kk].prob_combined() else kk
                            to_del.append(idx_to_delete)

                # make unique and sort
                to_del = sorted(list(set(to_del)), reverse=True)

                # delete
                for idx in to_del:
                    structured_elements.pop(idx)

    output: List[ElementGroup] = []
    for class_value, structured_elements in candidates_by_class.items():
        if class_value in detection_schema_by_class:
            output += structured_elements

    make_unique = {key: item for (key, item) in detection_schema_by_class.items() if item.takeBestInProximity}
    
    if len(make_unique.keys()) > 1:
        
        # make final selection based on distance of statements to each other
        # determine text distance of one statement to all others
        distances = []

        for k in range(0, len(output) - 1):
            for kk in range(k + 1, len(output)):
                distance = text_distance_table_groups(output[k], output[kk], document.elements)
                distances.append((distance, k, kk))

        # sort
        distances = list(sorted(distances, key=lambda x: x[0]))

        # combine one by one with closest distance
        final_groups = [{"dist": 0, "prob_score": 0, "indices": {key: None for key in make_unique.keys()}}]
        for dist_tuple in distances:
            # only combine if max distance is respected (or location was user defined -> prob == 2)
            if (dist_tuple[0] < MAX_DISTANCE_UNIQUE_PROXIMITY) or (output[dist_tuple[1]].prob_combined() == 2 and output[dist_tuple[2]].prob_combined() == 2):
                placed_item = False
                el1 = output[dist_tuple[1]]
                el2 = output[dist_tuple[2]]
                for final_group in final_groups:
                    # check if element can be added to group
                    if el1.detected_class != el2.detected_class and (final_group["indices"][el1.detected_class] is None or final_group["indices"][el1.detected_class] == dist_tuple[1]) and (
                            final_group["indices"][el2.detected_class] is None or final_group["indices"][el2.detected_class] == dist_tuple[2]):
                        final_group["indices"][el1.detected_class] = dist_tuple[1]
                        final_group["indices"][el2.detected_class] = dist_tuple[2]
                        final_group["dist"] += dist_tuple[0]
                        placed_item = True

                # create new group if item was not placed yet
                if not placed_item:
                    final_groups.append({"dist": 0, "prob_score": 0, "indices": {key: None for key in make_unique.keys()}})
                    final_groups[-1]["indices"][el1.detected_class] = dist_tuple[1]
                    final_groups[-1]["indices"][el2.detected_class] = dist_tuple[2]
                    final_groups[-1]["dist"] += dist_tuple[0]
            else:
                break

        # take the first group that has no None values
        groups_filtered = list(sorted([x for x in final_groups if None not in x["indices"].values()], key=lambda x: x['dist']))
        # compile matching score of final groups
        for g in groups_filtered:
            g['prob_score'] = 0
            for key in make_unique.keys():
                g['prob_score'] += output[g["indices"][key]].prob_combined()

        if len(groups_filtered) == 0:
            return []

        # take highest probabilities, then shortest distance
        groups_filtered = list(sorted(groups_filtered, key=lambda x: (-x['prob_score'], x['dist'])))
        group_chosen = groups_filtered[0]

        # delete values
        all_valid_indices = [group_chosen["indices"][key] for key in make_unique.keys()]
        for k in range(len(output) - 1, -1, -1):
            if k not in all_valid_indices:
                output.pop(k)
    
    return output


def get_structured_tables_from_locations(job_template: JobTemplate, document: StandardDocumentFormat, locations: List[ParseeLocation]) -> List[FinalOutputTableColumn]:

    element_groups = assemble(job_template, document, locations)

    # make final structured values
    output_values: List[FinalOutputTableColumn] = []
    for group in element_groups:
        output_values += group.structured_values(document.elements)

    return output_values


def final_tables_from_columns(columns: List[FinalOutputTableColumn]) -> List[FinalOutputTable]:

    by_identifier = {}

    for col in columns:
        table_id = (col.li_identifier, col.detected_class)
        if table_id not in by_identifier:
            by_identifier[table_id] = FinalOutputTable(col.detected_class, [col], col.li_identifier)
        else:
            by_identifier[table_id].columns.append(col)

    return list(by_identifier.values())



from typing import *
from decimal import Decimal
from copy import deepcopy

from parsee.storage.interfaces import StorageManager
from parsee.storage.vector_stores.simple_numpy import SimpleNumpyStore
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.templates.job_template import JobTemplate
from parsee.extraction.extractor_dataclasses import AssignedMeta, AssignedLocation, AssignedAnswer, AssignedBucket
from parsee.converters.image_creation import DiskImageCreator, ImageCreator


class InMemoryStorageManager(StorageManager):
    
    truth_questions: List[AssignedAnswer]
    truth_locations: List[AssignedLocation]
    truth_meta: List[AssignedMeta]
    truth_mappings: List[AssignedBucket]

    def __init__(self, available_models: Optional[List[MlModelSpecification]], custom_image_creator: Optional[ImageCreator] = None):
        super().__init__(SimpleNumpyStore(), DiskImageCreator() if custom_image_creator is None else custom_image_creator)
        self.models = available_models if available_models is not None else []
        self.truth_questions = []
        self.truth_locations = []

    def get_available_models(self) -> List[MlModelSpecification]:
        return self.models if self.models is not None else []

    def log_expense(self, service: str, amount: Decimal, class_id: str):
        print("expense", service, amount, class_id)
        
    def assign_truth_values(self, general_questions: Optional[List[AssignedAnswer]], locations: Optional[List[AssignedLocation]]):
        if general_questions is not None:
            self.truth_questions = general_questions
        if locations is not None:
            self.truth_locations = locations
        
    def db_values_template(self, job_template: JobTemplate, rerun_if_no_input: bool) -> JobTemplate:

        # adjust template
        template_new = deepcopy(job_template)

        # add assigned values
        template_new.detection.settings = {**template_new.detection.settings, "assigned": self.truth_locations}
        template_new.questions.settings = {**template_new.questions.settings, "assigned": self.truth_questions}

        # adjust models
        for item in template_new.detection.items:
            item.model = "assigned"
            item.mappingModel = "assigned"
        for item in template_new.meta:
            item.model = "assigned"
        for item in template_new.questions.items:
            item.model = "assigned"

        return template_new



from typing import *
from dataclasses import dataclass
from decimal import Decimal

from parsee.extraction.extractor_elements import FileReference


@dataclass
class Message:
    text: str
    references: List[FileReference]
    author: Optional[str] = None
    cost: Optional[Decimal] = None

    def __str__(self):
        if self.author is None:
            return self.text
        return f"{self.author}: {self.text}"

    def __repr__(self):
        return str(self)



from parsee.utils.helper import get_source_identifier
from parsee.storage.interfaces import *
from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.converters.main import load_document


class LocalFileManager(DocumentManager):

    def __init__(self, storage: StorageManager, document_paths: List[str]):
        super().__init__(storage)
        self.source_identifier_to_paths = {}
        for path in document_paths:
            self.source_identifier_to_paths[get_source_identifier(path)] = path

    def add_files(self, document_paths: List[str]):
        for path in document_paths:
            self.source_identifier_to_paths[get_source_identifier(path)] = path

    def load_with_source_identifier(self, source_identifier: str) -> StandardDocumentFormat:
        if source_identifier not in self.source_identifier_to_paths:
            raise Exception("file not found")
        return load_document(self.source_identifier_to_paths[source_identifier])

    def load_documents(self, references: List[FileReference], multimodal: bool, search_term: Optional[str], max_images: Optional[int], max_tokens: Optional[int], show_chunk_index: bool = False) -> Union[str, List[Base64Image]]:
        return self._load_documents(references, multimodal, search_term, max_images, max_tokens, self.load_with_source_identifier, show_chunk_index)



from functools import lru_cache
from typing import List, Tuple
from decimal import Decimal

from mistralai import Mistral, SDKError
import tiktoken
from tenacity import retry, stop_after_attempt, wait_random_exponential, retry_if_exception, after_log

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.extractor_dataclasses import Base64Image
from parsee.settings import chat_settings
import logging

logger = logging.getLogger(__name__)


class MistralModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer
        self.client = Mistral(api_key=model.api_key) if model.api_key is not None else None

    @retry(reraise=True,
           stop=stop_after_attempt(chat_settings.retry_attempts),
           retry=retry_if_exception(lambda x: isinstance(x, SDKError) and x.status_code == 429),
           wait=wait_random_exponential(multiplier=chat_settings.retry_wait_multiplier,
                                 min=chat_settings.retry_wait_min,
                                 max=chat_settings.retry_wait_max),
           after=after_log(logger, logging.DEBUG) )
    def _call_api(self, prompt: str, images: List[Base64Image]) -> Tuple[str, Decimal]:

        user_message_content = [
            {
                "type": "text",
                "text": prompt
            }
        ]
        if len(images) > 0:
            user_message_content += [
                {
                    "type": "image_url",
                    "image_url": f"data:{x.media_type};base64,{x.data}",
                }
                for x in images
            ]

        messages = [{"role": "user", "content": user_message_content}]
        if self.spec.system_message is not None:
            messages.insert(0, {"role": "system", "content": [{"type": "text", "text": self.spec.system_message}]})

        chat_response = self.client.chat.complete(
            model=self.spec.internal_name,
            messages=messages,
            temperature=0
        )

        answer = chat_response.choices[0].message.content
        cost_input = (int(chat_response.usage.prompt_tokens) * Decimal(self.spec.price_per_1k_tokens / 1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (int(chat_response.usage.completion_tokens) * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        cost_images = (len(images) * Decimal(self.spec.price_per_image)) if self.spec.price_per_image is not None else Decimal(0)
        final_cost = cost_input + cost_output + cost_images
        return answer, final_cost

    @lru_cache(maxsize=chat_settings.max_cache_size)
    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt, prompt.available_data if self.spec.multimodal else [])



from typing import *
import tempfile
import base64
import shutil
import os

import cv2
from numpy import ndarray, frombuffer
import numpy as np

from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl, ExtractedSource
from parsee.utils.enums import DocumentType
from pdf_reader.helper import make_images_from_pdf
from pdf_reader.converter import is_image
from parsee.extraction.extractor_dataclasses import Base64Image


def resize(image_cv2: ndarray, max_image_size: Optional[int]) -> ndarray:
    height, width, channels = image_cv2.shape
    if not (height > max_image_size or width > max_image_size):
        return image_cv2
    bigger_val = max(width, height)
    scale_factor = max_image_size / bigger_val
    algo = cv2.INTER_AREA if scale_factor < 1 else cv2.INTER_CUBIC
    img_resized = cv2.resize(image_cv2, (0, 0), fx=scale_factor, fy=scale_factor, interpolation=algo)
    return img_resized


def from_bytes(file_content: bytes, max_image_size: int) -> Base64Image:
    # open and resize image if necessary
    jpg_as_np = frombuffer(file_content, dtype=np.uint8)
    img = cv2.imdecode(jpg_as_np, cv2.IMREAD_COLOR)
    # resize
    img = resize(img, max_image_size)
    return from_numpy(img)


def from_numpy(numpy_img: ndarray) -> Base64Image:
    retval, buffer = cv2.imencode('.jpg', numpy_img)
    encoded_string = base64.b64encode(buffer).decode("utf-8")
    return Base64Image("image/jpeg", encoded_string)


def get_media_type_simple(file_path: str) -> str:
    if file_path.endswith(".jpg") or file_path.endswith(".jpeg"):
        return "image/jpeg"
    elif file_path.endswith(".png"):
        return "image/png"
    else:
        raise Exception("unsupported image type")


def from_file_paths(file_paths: List[str], max_image_size: int) -> List[Base64Image]:
    output = []
    for fp in file_paths:
        with open(fp, "rb") as f:
            output.append(from_bytes(f.read(), max_image_size))

    return output


class ImageCreator:

    def get_images(self, document: StandardDocumentFormat, element_selection: List[Union[ExtractedEl, ExtractedSource]], max_images: Optional[int], max_image_size: Optional[int]) -> List[Base64Image]:
        raise NotImplementedError


class DiskImageCreator(ImageCreator):

    def get_images(self, document: StandardDocumentFormat, element_selection: List[Union[ExtractedEl, ExtractedSource]], max_images: Optional[int], max_image_size: Optional[int]) -> List[Base64Image]:

        max_image_size = 2000 if max_image_size is None else max_image_size

        if document.file_path is None:
            return []

        output = []

        if document.source_type == DocumentType.PDF:
            if document.file_path is not None and is_image(document.file_path):
                output += from_file_paths([document.file_path], max_image_size)
            else:
                # collect all relevant pages
                page_indexes = []
                for el in element_selection:
                    if isinstance(el, ExtractedEl):
                        source = el.source
                    elif isinstance(el, ExtractedSource):
                        source = el
                    else:
                        raise Exception("unknown element type")
                    if source.other_info is not None and "page_idx" in source.other_info and int(source.other_info["page_idx"]) not in page_indexes and (max_images is None or len(page_indexes) < max_images):
                        page_indexes.append(int(source.other_info["page_idx"]))
                temp_dir = tempfile.TemporaryDirectory()
                images = make_images_from_pdf(document.file_path, temp_dir.name, [max_image_size], None)
                file_paths = [images[max_image_size][x] for x in page_indexes]
                output += from_file_paths(file_paths, max_image_size)
                # delete temp dir
                shutil.rmtree(temp_dir.name)
        else:
            raise Exception("unsupported document type, images can only be created from PDFs. To create a PDF from a HTML file, use a tool like pdfkit, then run pdfkit.from_file('your_file') and use the output file instead of the HTML.")

        return output


class DiskImageReader(ImageCreator):

    def __init__(self, images_dir: str):
        self.images_dir = images_dir

    def get_images(self, document: StandardDocumentFormat, element_selection: List[Union[ExtractedEl, ExtractedSource]], max_images: Optional[int], max_image_size: Optional[int]) -> List[Base64Image]:

        if document.source_type == DocumentType.PDF:
            if document.file_path is not None and is_image(document.file_path):
                page_indexes = [0]
            else:
                # collect all relevant pages
                page_indexes = []
                for el in element_selection:
                    if isinstance(el, ExtractedEl):
                        source = el.source
                    elif isinstance(el, ExtractedSource):
                        source = el
                    else:
                        raise Exception("unknown element type")
                    if source.other_info is not None and "page_idx" in source.other_info and int(source.other_info["page_idx"]) not in page_indexes and (
                            max_images is None or len(page_indexes) < max_images):
                        page_indexes.append(int(source.other_info["page_idx"]))
            # check that all pages can be found
            file_paths = []
            for page_idx in page_indexes:
                file_path = os.path.join(self.images_dir, f"{document.source_identifier}_p{page_idx}.jpg")
                if os.path.exists(file_path):
                    file_paths.append(file_path)
                else:
                    raise Exception(f"file not found, expected location: {file_path}")
            return from_file_paths(file_paths, max_image_size)
        else:
            raise Exception("unsupported document type, images can only be created from PDFs. To create a PDF from a HTML file, use a tool like pdfkit, then run pdfkit.from_file('your_file') and use the output file instead of the HTML.")